'use strict';

Object.defineProperty(exports, "__esModule", {
  value: true
});
exports.default = stringifySqlAST;

var _util = require('../util');

function stringifySqlAST(topNode, context) {
  (0, _util.validateSqlAST)(topNode);

  let { selections, joins, wheres } = _stringifySqlAST(null, topNode, '', context, [], [], []);

  selections = [...new Set(selections)];

  if (!selections.length) return '';

  let sql = 'SELECT\n  ' + selections.join(',\n  ') + '\n' + joins.join('\n');
  if (wheres.length) {
    sql += '\nWHERE ' + wheres.join(' AND ');
  }
  return sql;
}

function _stringifySqlAST(parent, node, prefix, context, selections, joins, wheres) {
  switch (node.type) {
    case 'table':
      if (node.where) {
        const whereCondition = node.where(`${ quote(node.as) }`, node.args || {}, context);
        if (whereCondition) {
          wheres.push(`${ whereCondition }`);
        }
      }

      if (node.sqlJoin) {
        const joinCondition = node.sqlJoin(`${ quote(parent.as) }`, `${ quote(node.as) }`);

        joins.push(`LEFT JOIN ${ node.name } AS ${ quote(node.as) } ON ${ joinCondition }`);
      } else if (node.joinTable) {
        if (!node.sqlJoins) throw new Error('Must set "sqlJoins" for a join table.');
        const joinCondition1 = node.sqlJoins[0](`${ quote(parent.as) }`, `${ quote(node.joinTableAs) }`);
        const joinCondition2 = node.sqlJoins[1](`${ quote(node.joinTableAs) }`, `${ quote(node.as) }`);

        joins.push(`LEFT JOIN ${ node.joinTable } AS ${ quote(node.joinTableAs) } ON ${ joinCondition1 }`, `LEFT JOIN ${ node.name } AS ${ quote(node.as) } ON ${ joinCondition2 }`);
      } else {
        joins.push(`FROM ${ node.name } AS ${ quote(node.as) }`);
      }

      for (let child of node.children) {
        _stringifySqlAST(node, child, parent ? prefix + node.as + '__' : prefix, context, selections, joins, wheres);
      }

      break;
    case 'column':
      selections.push(`${ quote(parent.as) }.${ quote(node.name) } AS ${ quote(prefix + node.as) }`);
      break;
    case 'columnDeps':
      for (let name in node.names) {
        selections.push(`${ quote(parent.as) }.${ quote(name) } AS ${ quote(prefix + node.names[name]) }`);
      }
      break;
    case 'composite':
      const keys = node.name.map(key => `${ quote(parent.as) }.${ quote(key) }`);

      selections.push(`CONCAT(${ keys.join(', ') }) AS ${ quote(prefix + node.fieldName) }`);
      break;
    case 'noop':
      return;
    default:
      throw new Error('unexpected/unknown node type reached: ' + (0, _util.inspect)(node));
  }
  return { selections, joins, wheres };
}

function quote(str) {
  return '`' + str + '`';
}